import { AutoMap } from '@automapper/classes';
import {
  IsOptional,
  IsBoolean,
  IsInt,
  IsEnum,
  ValidateNested,
  ArrayMinSize,
  IsUUID,
  IsArray,
  IsISO8601,
} from 'class-validator';
import { Transform, Type } from 'class-transformer';
import { ScheduleDTO } from './schedule.dto';
import { MarginDurationsDTO } from './margin-durations.dto';
import { WaypointDTO } from './waypoint.dto';
import { intToFrequency } from './validators/frequency.mapping';
import { IsSchedule } from './validators/decorators/is-schedule.validator';
import { HasValidPositionIndexes } from './validators/decorators/valid-position-indexes.validator';
import { Frequency } from '@modules/ad/core/ad.types';

export class CreateAdRequestDTO {
  @IsUUID(4)
  @AutoMap()
  userId: string;

  @IsOptional()
  @IsBoolean()
  @AutoMap()
  driver?: boolean;

  @IsOptional()
  @IsBoolean()
  @AutoMap()
  passenger?: boolean;

  @Transform(({ value }) => intToFrequency(value), {
    toClassOnly: true,
  })
  @IsEnum(Frequency)
  @AutoMap()
  frequency: Frequency;

  @IsISO8601({
    strict: true,
    strictSeparator: true,
  })
  @AutoMap()
  fromDate: string;

  @IsISO8601({
    strict: true,
    strictSeparator: true,
  })
  @AutoMap()
  toDate: string;

  @Type(() => ScheduleDTO)
  @IsSchedule()
  @ValidateNested({ each: true })
  @AutoMap()
  schedule: ScheduleDTO;

  @IsOptional()
  @Type(() => MarginDurationsDTO)
  @ValidateNested({ each: true })
  @AutoMap()
  marginDurations?: MarginDurationsDTO;

  @IsOptional()
  @IsInt()
  @AutoMap()
  seatsProposed?: number;

  @IsOptional()
  @IsInt()
  @AutoMap()
  seatsRequested?: number;

  @IsOptional()
  @IsBoolean()
  @AutoMap()
  strict?: boolean;

  @IsArray()
  @ArrayMinSize(2)
  @HasValidPositionIndexes()
  @ValidateNested({ each: true })
  @AutoMap()
  waypoints: WaypointDTO[];
}
