import { Test, TestingModule } from '@nestjs/testing';
import { CreateAdUseCase } from '../../../core/usecases/create-ad.usecase';
import { CreateAdRequestDTO } from '../../../interface/commands/create-ad.request.dto';
import { AdsRepository } from '../../../adapters/secondaries/ads.repository';
import { CreateAdCommand } from '../../../interface/commands/create-ad.command';
import { AutomapperModule } from '@automapper/nestjs';
import { classes } from '@automapper/classes';
import { Frequency } from '../../../interface/commands/frequency.enum';
import { Ad } from '../../../core/entities/ad';
import { AdProfile } from '../../../mappers/ad.profile';
import { Waypoint } from '../../../interface/commands/waypoint';
import { AdDTO } from '../../../core/dtos/ad.dto';
import { PARAMS_PROVIDER } from '../../../ad.di-tokens';
import { MESSAGE_PUBLISHER } from '../../../../../app.constants';

const originWaypoint: Waypoint = {
  position: 0,
  lon: 48.68944505415954,
  lat: 6.176510296462267,
  houseNumber: '5',
  street: 'Avenue Foch',
  locality: 'Nancy',
  postalCode: '54000',
  country: 'France',
};
const destinationWaypoint: Waypoint = {
  position: 1,
  lon: 48.8566,
  lat: 2.3522,
  locality: 'Paris',
  postalCode: '75000',
  country: 'France',
};
const originWaypointWithoutPosition: Waypoint = {
  lon: 43.2965,
  lat: 5.3698,
  locality: 'Marseille',
  postalCode: '13000',
  country: 'France',
};
const destinationWaypointWithoutPosition: Waypoint = {
  lon: 43.7102,
  lat: 7.262,
  locality: 'Nice',
  postalCode: '06000',
  country: 'France',
};
const newAdRequest: CreateAdRequestDTO = {
  userUuid: '113e0000-0000-4000-a000-000000000000',
  driver: true,
  passenger: false,
  frequency: Frequency.RECURRENT,
  fromDate: new Date('2023-05-01'),
  toDate: new Date('2023-08-20'),
  schedule: {
    tue: '08:00',
    wed: '08:30',
  },
  marginDurations: {
    mon: undefined,
    tue: undefined,
    wed: undefined,
    thu: undefined,
    fri: undefined,
    sat: undefined,
    sun: undefined,
  },
  seatsProposed: 2,
  waypoints: [originWaypoint, destinationWaypoint],
};

const mockMessagePublisher = {
  publish: jest.fn().mockImplementation(),
};
const mockDefaultParamsProvider = {
  getParams: () => {
    return {
      MON_MARGIN: 900,
      TUE_MARGIN: 900,
      WED_MARGIN: 900,
      THU_MARGIN: 900,
      FRI_MARGIN: 900,
      SAT_MARGIN: 900,
      SUN_MARGIN: 900,
      DRIVER: false,
      SEATS_PROPOSED: 3,
      PASSENGER: true,
      SEATS_REQUESTED: 1,
      STRICT: false,
    };
  },
};
const mockAdRepository = {
  create: jest
    .fn()
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    .mockImplementationOnce((command?: CreateAdCommand) => {
      return Promise.resolve({
        ...newAdRequest,
        uuid: 'ad000000-0000-4000-a000-000000000000',
        createdAt: new Date('2023-05-01'),
      });
    })
    .mockImplementationOnce(() => {
      throw new Error('Already exists');
    })
    .mockImplementation(),
};
describe('CreateAdUseCase', () => {
  let createAdUseCase: CreateAdUseCase;
  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [AutomapperModule.forRoot({ strategyInitializer: classes() })],
      providers: [
        {
          provide: AdsRepository,
          useValue: mockAdRepository,
        },
        {
          provide: MESSAGE_PUBLISHER,
          useValue: mockMessagePublisher,
        },
        CreateAdUseCase,
        AdProfile,
        {
          provide: PARAMS_PROVIDER,
          useValue: mockDefaultParamsProvider,
        },
      ],
    }).compile();

    createAdUseCase = module.get<CreateAdUseCase>(CreateAdUseCase);
  });
  it('should be defined', () => {
    expect(createAdUseCase).toBeDefined();
  });
  describe('execution', () => {
    const newAdCommand = new CreateAdCommand(newAdRequest);
    it('should create a new ad', async () => {
      const newAd: Ad = await createAdUseCase.execute(newAdCommand);
      expect(newAd.userUuid).toBe('113e0000-0000-4000-a000-000000000000');
      expect(newAd.uuid).toBe('ad000000-0000-4000-a000-000000000000');
    });
    it('should throw an error if the ad already exists', async () => {
      await expect(
        createAdUseCase.execute(newAdCommand),
      ).rejects.toBeInstanceOf(Error);
    });
  });

  describe('Ad parameter default setting ', () => {
    beforeEach(() => {
      mockAdRepository.create.mockClear();
    });

    it('should define minimal recurrent ad as passenger', async () => {
      const minimalRecurrentAdRequest: CreateAdRequestDTO = {
        userUuid: '224e0000-0000-4000-a000-000000000000',
        frequency: Frequency.RECURRENT,
        fromDate: new Date('2023-05-01'),
        toDate: new Date('2024-05-01'),
        schedule: {
          mon: '08:00',
        },
        waypoints: [originWaypoint, destinationWaypoint],
      };
      const newAdCommand = new CreateAdCommand(minimalRecurrentAdRequest);
      const expectedAdCreation: AdDTO = {
        userUuid: '224e0000-0000-4000-a000-000000000000',
        frequency: Frequency.RECURRENT,
        fromDate: new Date('2023-05-01'),
        toDate: new Date('2024-05-01'),
        monTime: '08:00',
        tueTime: undefined,
        wedTime: undefined,
        thuTime: undefined,
        friTime: undefined,
        satTime: undefined,
        sunTime: undefined,
        monMargin: 900,
        tueMargin: 900,
        wedMargin: 900,
        thuMargin: 900,
        friMargin: 900,
        satMargin: 900,
        sunMargin: 900,
        driver: false,
        seatsProposed: 3,
        passenger: true,
        seatsRequested: 1,
        strict: false,
        waypoints: {
          create: [
            {
              uuid: undefined,
              adUuid: undefined,
              position: 0,
              lon: 48.68944505415954,
              lat: 6.176510296462267,
              houseNumber: '5',
              street: 'Avenue Foch',
              locality: 'Nancy',
              postalCode: '54000',
              country: 'France',
            },
            {
              uuid: undefined,
              adUuid: undefined,
              position: 1,
              lon: 48.8566,
              lat: 2.3522,
              locality: 'Paris',
              postalCode: '75000',
              country: 'France',
            },
          ],
        },
        createdAt: undefined,
      };
      await createAdUseCase.execute(newAdCommand);
      expect(mockAdRepository.create).toBeCalledWith(expectedAdCreation, {
        waypoints: true,
      });
    });
    it('should create a passenger ad with waypoints without position', async () => {
      const newPunctualPassengerAdRequest: CreateAdRequestDTO = {
        userUuid: '113e0000-0000-4000-a000-000000000000',
        passenger: true,
        frequency: Frequency.PUNCTUAL,
        departureDate: new Date('2023-05-22 09:36'),
        seatsRequested: 1,
        waypoints: [
          originWaypointWithoutPosition,
          destinationWaypointWithoutPosition,
        ],
        schedule: {},
      };
      const newAdCommand = new CreateAdCommand(newPunctualPassengerAdRequest);
      const expectedAdCreation: AdDTO = {
        userUuid: '113e0000-0000-4000-a000-000000000000',
        frequency: Frequency.PUNCTUAL,
        fromDate: new Date('2023-05-22'),
        toDate: new Date('2023-05-22'),
        monTime: '09:36',
        tueTime: undefined,
        wedTime: undefined,
        thuTime: undefined,
        friTime: undefined,
        satTime: undefined,
        sunTime: undefined,
        monMargin: 900,
        tueMargin: 900,
        wedMargin: 900,
        thuMargin: 900,
        friMargin: 900,
        satMargin: 900,
        sunMargin: 900,
        driver: false,
        seatsProposed: 3,
        passenger: true,
        seatsRequested: 1,
        strict: false,
        waypoints: {
          create: [
            {
              uuid: undefined,
              adUuid: undefined,
              position: 0,
              lon: 43.2965,
              lat: 5.3698,
              locality: 'Marseille',
              postalCode: '13000',
              country: 'France',
            },
            {
              uuid: undefined,
              adUuid: undefined,
              position: 1,
              lon: 43.7102,
              lat: 7.262,
              locality: 'Nice',
              postalCode: '06000',
              country: 'France',
            },
          ],
        },
        createdAt: undefined,
      };
      await createAdUseCase.execute(newAdCommand);
      expect(mockAdRepository.create).toBeCalledWith(expectedAdCreation, {
        waypoints: true,
      });
    });
  });
});
